@@ -9,33 +9,86 @@ import json
 import itertools
 import operator
 from operator import itemgetter
+from collections import Counter
 
 import numpy as np
 
+from matplotlib import mpl
 import matplotlib.pyplot as plt
 
-import ImageContour
-
 #from list_utils:
-from np_utils import ( totuple, interpGen, flatten, ziptranspose, roll,
-                       deletecases,partition, polyCirculationDirection,
-                       groupByFunction, getElementConnections,
-                       getChainsFromConnections, removeDuplicates,
-                       removeAdjacentDuplicates )
+from np_utils import ( totuple, interp, interpGen, flatten, ziptranspose,
+                       roll, removeDuplicates, removeAdjacentDuplicates,
+                       deletecases, partition, groupByFunction,
+                       getElementConnections, getChainsFromConnections, )
+#from gen_utils
+from np_utils import minmax
 #from func_utils
-from np_utils import compose
+from np_utils import compose,kwdPop
 #from np_utils:
-from np_utils import ( limitInteriorPoints, limitInteriorPointsInterpolating,
+from np_utils import ( polyArea, polyPerimeter, polyCentroid,
+                       polyCirculationDirection,pointDistance,
+                       limitInteriorPoints, limitInteriorPointsInterpolating,
                        addBorder, getValuesAroundPointInArray )
 
+import ImageContour
+from temperatureCmaps import temperatureCmap,temperatureDarkCmap,temperatureLightCmap
+
 def GetValuesAroundSCPoint(watershed2d,point,wrapX=False,wrapY=False):
     vals = getValuesAroundPointInArray(watershed2d,point,wrapX,wrapY)
     return ( tuple(vals.tolist()) if vals!=None else (None,None,None) )
 
-def _kwdPop(kwds,key,defaultValue):
-    '''If a dictionary has a key, pop the value and return it,
-       otherwise return defaultValue'''
-    return ( kwds.pop(key) if key in kwds else defaultValue )
+def BreakLinesIntoEvenPieces(points,nSegments):
+    '''Take the points from a multi-line and re-interpolate the
+       points to get nSegments evenly spaced segments'''
+    nSegments = (1 if nSegments<1 else nSegments)
+    ptsArr = np.array(points)
+    nodalLengths = np.concatenate([ [0], np.cumsum(pointDistance(ptsArr[:-1],ptsArr[1:])) ])
+    totalLength = nodalLengths[-1]
+    splitPoints = np.arange(0,totalLength+1e-10,totalLength/nSegments)[1:-1]
+    indsBefore = np.array([ np.where(nodalLengths<=s)[0][-1] for s in splitPoints ])
+    if len(indsBefore)>0:
+        interpBefore,interpAfter = nodalLengths[indsBefore,],nodalLengths[indsBefore+1,]
+        floatPart = (splitPoints-interpBefore)/(interpAfter-interpBefore)
+        interpIndices = [0] + (indsBefore+floatPart).tolist() + [-1]
+    else:
+        interpIndices = [0,-1]
+    return [ interp(ptsArr,i) for i in interpIndices ]
+
+def _getReportPixel(arr):
+    '''Get a function that can be passed to the 'format_coord' method of a matplotlib axis'''
+    def report_pixel(x,y):
+        x,y,s = int(x+0.5),int(y+0.5),arr.shape
+        return ( "value=" + str(arr[y,x]) + "  x=" + str(x) + " y=" + str(y)
+                 if 0<x<s[1] and 0<y<s[0] else
+                 "x=" + str(x) + " y=" + str(y) )
+    return report_pixel
+
+def _prepMinMaxForColoredPlot(vals,MinMax=None):
+    if MinMax==None:
+        mmin,mmax = minmax(deletecases(vals,[None]))
+        MinMax = minmax([mmin,-mmin,mmax,-mmax])
+    return MinMax
+
+def _prepColorsForColoredPlot(vals,kwds,MinMax=None):
+    cmap = kwdPop( kwds, 'cmap', temperatureCmap )
+    Min,Max = _prepMinMaxForColoredPlot(vals,MinMax)
+    norm = mpl.colors.Normalize(Min, Max)
+    scalarMap = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
+    colorList = [ ( None if v==None else
+                    scalarMap.to_rgba(v) )
+                 for v in vals ]
+    return colorList
+
+def drawColorbar(vals,MinMax=None,**kwds):
+    '''Use the same setup as _prepColorsForColoredPlot to instead make a
+       matplotlib colorbar'''
+    kwds['cmap'] = kwdPop( kwds, 'cmap', temperatureCmap ) # sets default value
+    Min,Max = _prepMinMaxForColoredPlot(vals,MinMax)
+    kwds['norm'] = mpl.colors.Normalize(Min, Max)
+    kwds['mappable'] = plt.cm.ScalarMappable(norm=kwds['norm'], cmap=kwds['cmap'])
+    kwds['mappable']._A = []
+    plt.colorbar(**kwds)
 
 class SubContour(object):
     '''A class to hold the data for a single SubContour (basically a connected list of points)
@@ -72,10 +125,15 @@ class SubContour(object):
         '''Returns connection values at the endpoints that are not one of the 2 main values'''
         return tuple(sorted( set(self.startPointValues+self.endPointValues).difference(self.values) ))
     
+    def getLength(self):
+        '''Compute the sum of the distances between the points'''
+        return pointDistance(self.points[:-1],self.points[1:]).sum()
+    
     def plot(self,*args,**kwds):
         x = [ p[0]-0.5 for p in self.points ]
         y = [ p[1]-0.5 for p in self.points ]
         return plt.plot( y,x, *args, **kwds )
+    
     def plotT(self,*args,**kwds):
         x = [ p[0]-0.5 for p in self.points ]
         y = [ p[1]-0.5 for p in self.points ]
@@ -92,6 +150,33 @@ class QuadPoint(object):
         self.values=values
         self.point=point
 
+def _getNumUniqueValues(values,bgVals):
+    '''Get the number of unique values connected to this junction.
+       All background values are considered the same'''
+    v,b = set(values),set(bgVals)
+    return len(v.difference(b)) + (1 if v.intersection(b) else 0)
+
+class Junction(object):
+    '''A class to hold information about a junction:
+       point: xy location
+       subContourIDs: integer ID's (indices) for connected subContours
+       cellIDs: integer ID's for connected cells
+       nConnections: number of connected subContours'''
+    point = None
+    subContourIDs = None
+    cellIDs = None
+    nConnections = None
+    def __init__(self,point,subContourIDs,cellIDs):
+        self.point = tuple(point)
+        self.subContourIDs = tuple(sorted(set(subContourIDs)))
+        self.cellIDs = tuple(sorted(set(cellIDs)))
+        self.nConnections = len(self.subContourIDs)
+    
+    def getNCellConnections(self,bgVals):
+        '''Get the number of unique cell ID's connected to this junction.
+           All background values are considered the same'''
+        return _getNumUniqueValues(self.cellIDs,bgVals)
+
 class DegenerateNetworkException(Exception):
     pass
 
@@ -106,10 +191,11 @@ class CellNetwork(object):
                                 # (<index into subContours>,<boolean that determines if this is forward (True) or backwards (False)>)
                                 # Don't use this directly, use the GetContourPoints method instead
                                 # no more negative values... (used to mean reverse the contour when inserting)
-    allValues = []
+    allValues = [] # list of all real CellIDs
+    # backgroundValues = [] # list of all background CellIDs
     def __init__(self,**kwds):
         '''Create a CellNetwork
-           All arguments are converted to attributes; all are optionally,
+           All arguments are converted to attributes; all are optional,
            but these are needed for functions to work:
              * subContours: list of SubContour objects
              * contourOrderingByValue: dictionary with keys that are cellIDs and tuple values like:
@@ -117,23 +203,59 @@ class CellNetwork(object):
                                          <boolean that determines if this is forward (True) or backwards (False)> )'''
         for k in kwds:
             self.__dict__[k] = kwds[k]
+        # if 'backgroundValues' not in kwds.keys():
+        #     self.backgroundValues = []
         if 'quadPoints' not in kwds.keys():
             if 'subContours' in kwds.keys():
                 self.UpdateQuadPoints()
     
+    def GetAllJunctions(self):
+        '''Get all the junctions in the network (points where sc's meet)'''
+        # Create a list of subContours ID's that are attached to each junction point
+        # And a list of cell ID values that are attached to each point
+        junctionSubContoursIDs = {}
+        junctionValues = {}
+        for i,sc in enumerate(self.subContours):
+            for endpt,values in [ (sc.points[0],sc.startPointValues),
+                                  (sc.points[-1],sc.endPointValues) ]:
+                pt = tuple(endpt)
+                junctionSubContoursIDs.setdefault(pt,set()).add(i)
+                junctionValues.setdefault(pt,set()).update(values)
+        allJunctionPoints = sorted(junctionSubContoursIDs.keys())
+        junctions = [ Junction(pt,scids,values)
+                     for pt in allJunctionPoints
+                     for scids in [junctionSubContoursIDs[pt]]
+                     for values in [junctionValues[pt]] ]
+        return junctions
+    
+    def GetJunctionsByNSubContours(self):
+        junctions = self.GetAllJunctions()
+        junctionsByNConns = {}
+        for j in junctions:
+            junctionsByNConns.setdefault(j.nConnections,[]).append(j)
+        return junctionsByNConns
+    
+    def GetJunctionsByNCells(self,bgVals):
+        junctions = self.GetAllJunctions()
+        junctionsByNConns = {}
+        for j in junctions:
+            junctionsByNConns.setdefault(j.getNCellConnections(bgVals),[]).append(j)
+        return junctionsByNConns
+    
     def UpdateQuadPoints(self):
         '''Update quadPoints from subContours
            State: Changes state of quadPoints variable, but only to be more consistent'''
-        quadPoints = sorted(set( [ (sc.startPointValues,tuple(sc.points[0])) for sc in self.subContours
-                                                                             if len(sc.startPointValues)==4 ] +
-                                 [ (sc.endPointValues,tuple(sc.points[-1])) for sc in self.subContours
-                                                                            if len(sc.endPointValues)==4 ] ))
+        startQuads = [ (sc.startPointValues,tuple(sc.points[0])) for sc in self.subContours
+                                                                 if len(sc.startPointValues)==4 ]
+        endQuads = [ (sc.endPointValues,tuple(sc.points[-1])) for sc in self.subContours
+                                                              if len(sc.endPointValues)==4 ]
+        quadPoints = sorted(set(startQuads+endQuads))
         self.quadPoints = [ QuadPoint(vals,pt) for vals,pt in quadPoints ]
+    
     def GetContourPoints(self,v,closeLoop=True):
         '''Get Contour points around a value v
            State: Access only'''
-        def reverseIfFalse(l,direction):
-            return ( l if direction else l[::-1] )
+        reverseIfFalse = lambda l,direction: ( l if direction else l[::-1] )
         scPointsList = [ reverseIfFalse( self.subContours[index].points, direction ) # reconstruct sc's, flipping if direction is False
                         for index,direction in self.contourOrderingByValue[v] ] # for each index & direction tuple
         contourPoints = [ totuple(pt) for scp in scPointsList for pt in scp[:-1] ] # start point is not end point; assumed to be cyclical...
@@ -169,14 +291,57 @@ class CellNetwork(object):
             polyList.append( [ v,indexList ] )
         return [xyList,polyList]
         
-    def UpdateAllValues(self):
+    def UpdateAllValues(self,bgVals=(0,1)):
         '''Go through the values in all the subContours and collect a list of all of them
            State: Changes state of allValues, but only to to be more consistent'''
         self.allValues = sorted(set( [ v for sc in self.subContours 
                                          for v in sc.values
-                                         if v!=1 ] ))
+                                         if v not in bgVals ] ))
+    def GetAllValues(self,update=True,bgVals=(0,1)):
+        '''Update all values and then return them.
+           Not efficient for multiple uses.
+           State: Changes the state of allValues when update=True, but only to be more consistent'''
+        if update:
+            self.UpdateAllValues(bgVals=bgVals)
+        return self.allValues
+    
+    def GetCellPerimeters(self,update=False,bgVals=(0,1)):
+        '''For each cell in allValues, get the cell perimeter
+           State: Changes the state of allValues when update=True, but only to be more consistent'''
+        if update:
+            self.UpdateAllValues(bgVals=bgVals)
+        return [ polyPerimeter(self.GetContourPoints(v))
+                for v in self.allValues ]
+    
+    def GetCellAreas(self,update=False,bgVals=(0,1)):
+        '''For each cell in allValues, get the cell area
+           State: Changes the state of allValues when update=True, but only to be more consistent'''
+        if update:
+            self.UpdateAllValues(bgVals=bgVals)
+        return [ polyArea(self.GetContourPoints(v,closeLoop=False))
+                for v in self.allValues ]
+    
+    def GetCellCentroids(self,update=False,bgVals=(0,1)):
+        '''For each cell in allValues, get the cell centroid
+           State: Changes the state of allValues when update=True, but only to be more consistent'''
+        if update:
+            self.UpdateAllValues(bgVals=bgVals)
+        return [ polyCentroid(self.GetContourPoints(v,closeLoop=False))
+                for v in self.allValues ]
+    
+    def GetSubContourLengths(self):
+        '''Compute the length of each subContour with sc.getLength'''
+        return [ sc.getLength() for sc in self.subContours ]
     
     def GetAllPoints(self):
+        '''Get a set of all points in the subContours (as 2-tuples)
+           Sorting to done based on the order points appear in the subContours
+           This ordering helps ensure that matched CN's translate properly to
+           matched cvd's
+           State: Access only'''
+        return removeDuplicates([ tuple(pt) for sc in self.subContours for pt in sc.points ])
+    
+    def GetAllPointsSorted(self):
         '''Get a sorted set of all points in the subContours
            State: Access only'''
         return sorted(set( [ tuple(pt) for sc in self.subContours for pt in sc.points ] ))
@@ -442,35 +607,130 @@ class CellNetwork(object):
     def scPlot(self,*args,**kwds):
         '''Plot the subContours in a way that can be overlaid on an imshow
            State: Access only'''
-        for sc in self.subContours:
-            _=sc.plot(*args,**kwds)
+        reverseXY = kwdPop( kwds, 'reverseXY', False )
+        if reverseXY:
+            return [ sc.plotT(*args,**kwds)
+                    for sc in self.subContours ]
+        else:
+            return [ sc.plot(*args,**kwds)
+                    for sc in self.subContours ]
+                
     
     def scPlotT(self,*args,**kwds):
         '''Plot the subContours in a way that can be overlaid on a transposed imshow
            (matches closely with diagramPlot in VFMLite)
            State: Access only'''
-        for sc in self.subContours:
-            _=sc.plotT(*args,**kwds)
+        kwds['reverseXY'] = True
+        return self.scPlot(*args,**kwds)
     
     def cellPlot(self,*args,**kwds):
         '''Plot the full contours in a way that can be overlaid on an imshow
+           Some default kwd args are not shown:
+             plotFunction=plt.plot
+             colorList=None
+             reverseXY=False
+           All other options are passed to the plotFunction (default plt.plot)
            State: Access only'''
-        plotFunction = _kwdPop( kwds, 'plotFunction', plt.plot )
-        reverseXY    = _kwdPop( kwds, 'reverseXY'   , False    )
+        plotFunction = kwdPop( kwds, 'plotFunction', plt.plot )
+        colorList = kwdPop( kwds, 'colorList', None )
+        reverseXY = kwdPop( kwds, 'reverseXY', False )
         
         contourPoints = { v:self.GetContourPoints(v) for v in self.contourOrderingByValue.keys() }
-        for v in contourPoints.keys():
+        for i,v in enumerate(sorted(contourPoints.keys())):
             x = [ p[0]-0.5 for p in contourPoints[v] ]
             y = [ p[1]-0.5 for p in contourPoints[v] ]
             x,y = (y,x) if reverseXY else (x,y)
+            if colorList!=None:
+                if colorList[i]==None: # skip cells with no color
+                    continue
+                kwds['color'] = colorList[i] # override the default color
             _=plotFunction( y,x, *args, **kwds )
     
     def cellPlotT(self,*args,**kwds):
         '''Plot the full contours in a way that can be overlaid on a transposed imshow
-           (matches closely with diagramPlot in VFMLite)
+           Calls cellPlot with reverseXY=True
            State: Access only'''
         kwds['reverseXY'] = True
         return self.cellPlot(*args,**kwds)
+    
+    def tensionPlot(self,tensions,*args,**kwds):
+        '''False color plot of tensions
+           Some kwd args with default values are not shown:
+             linewidth=2
+             tensionMinMax=None
+             reverseXY=False
+             blackOutlines=False
+             cmap=temperatureCmap
+           All other options are passed to cellPlot
+           State: Access only'''
+        linewidth = kwdPop( kwds, 'linewidth', 2 )
+        tensionMinMax = kwdPop( kwds, 'tensionMinMax', None )
+        blackOutlines = kwdPop( kwds, 'blackOutlines', False )
+        
+        colorList = _prepColorsForColoredPlot(tensions,kwds,tensionMinMax)
+        
+        if blackOutlines:
+            kwds['color'],kwds['linewidth'] = 'k',linewidth+1
+            _=self.scPlot(*args,**kwds)
+        
+        reverseXY = kwdPop( kwds, 'reverseXY', False ) # must do this after the self.scPlot...
+        
+        plotList = []
+        for color,sc in zip(colorList,self.subContours):
+            if color!=None:
+                scplot = ( sc.plotT if reverseXY else sc.plot )
+                kwds['color'],kwds['linewidth'] = color,linewidth
+                plotList.append(  scplot(*args,**kwds)  )
+        return plotList
+    
+    def tensionPlotT(self,tensions,*args,**kwds):
+        '''tensionPlot, but with reverseXY=True (flips x and y axes)
+           State: Access only'''
+        kwds['reverseXY'] = True
+        return self.tensionPlot(tensions,*args,**kwds)
+    
+    def pressurePlot(self,pressures,*args,**kwds):
+        '''False color plot of pressures
+           Some kwd args with default values are not shown:
+             plotFunction=plt.fill
+             pressureMinMax=None
+             cmap=temperatureCmap
+           All other options are passed to cellPlot
+           State: Access only'''
+        kwds['plotFunction'] = kwdPop( kwds, 'plotFunction', plt.fill )
+        pressureMinMax = kwdPop( kwds, 'pressureMinMax', None )
+        kwds['colorList'] = _prepColorsForColoredPlot(pressures,kwds,pressureMinMax)
+        
+        return self.cellPlot(*args,**kwds)
+    
+    def pressurePlotT(self,pressures,*args,**kwds):
+        '''pressurePlot, but with reverseXY=True (flips x and y axes)
+           State: Access only'''
+        kwds['reverseXY'] = True
+        return self.pressurePlot(pressures,*args,**kwds)
+        
+    def pressureImagePlot(self,pressures,waterArr,*args,**kwds):
+        kwds['cmap'] = kwdPop( kwds, 'cmap', temperatureCmap ) # sets default value
+        pressureMinMax = kwdPop( kwds, 'pressureMinMax', None )
+        pressureErrors = kwdPop( kwds, 'pressureErrors', [0]*len(pressures) )
+        pMin,pMax = _prepMinMaxForColoredPlot(pressures,pressureMinMax)
+        
+        rPlot = np.zeros(waterArr.shape,dtype=np.float)
+        for i,v in enumerate(sorted(self.contourOrderingByValue.keys())):
+            p,pErr = pressures[i],pressureErrors[i]
+            if p!=None:
+                wh = np.where(waterArr==v)
+                nPts = len(wh[0])
+                rPlot[wh] = ( np.random.normal(p,pErr,nPts) if pErr!=0 else
+                              [p]*nPts )
+        
+        plt.imshow(rPlot,cmap=temperatureCmap,vmin=pMin,vmax=pMax)
+        plt.gca().format_coord = _getReportPixel(rPlot) # enable pixel reporting
+    
+    def pressureImageSpecklePlot(self,pressures,pressureErrors,waterArr,*args,**kwds):
+        kwds['pressureErrors'] = pressureErrors
+        self.pressureImagePlot(pressures,waterArr,*args,**kwds)
+    
 
 def GetFlatDataFromCellNetwork(cn):
     '''Convert a CellNetwork with all its nested objects into flat datastructure
@@ -500,6 +760,21 @@ def GetCellNetworkFromFlatData(fdcn):
                             'contourOrderingByValue' : { int(k):v for k,v in fdcn['contourOrderingByValue'].iteritems() },
                             'allValues': fdcn['allValues'], } )
 
+def GetCellNetworkListStaticFromJsonFile(filename):
+    with open(filename) as fid:
+        cnListFlat = json.load(fid)
+    cnList = [ GetCellNetworkFromFlatData(fdcn)
+              for fdcn in cnListFlat ]
+    return cnList
+
+def GetCellNetworkListPrevNextFromJsonFile(filename):
+    with open(filename) as fid:
+        cnListPrevFlat,cnListNextFlat = json.load(fid)
+    cnListPrev,cnListNext = [ [ GetCellNetworkFromFlatData(fdcn)
+                               for fdcn in cnListFlat ]
+                             for cnListFlat in cnListPrevFlat,cnListNextFlat ]
+    return cnListPN
+
 def SubContourListfromCVLSList(cVLS_List,startPointValues_List=[],endPointValues_List=[]):
     '''Get a list of SubContour objects from an old list of cVLS's'''
     if startPointValues_List==[]:
@@ -585,6 +860,58 @@ def GetXYListAndPolyListFromCellNetworkList(cellNetworkList,closeLoops=True):
 def GetCVDListFromCellNetworkList(cellNetworkList):
     return [ cn.GetCellularVoronoiDiagram() for cn in cellNetworkList ]
 
+def GetDanglingBoundaryCells(cn,bgVals):
+    '''Get any boundary cells that are only connected to other boundary cells'''
+    bgSet = set(bgVals)
+    # For each sc, get the one or two connected cell ID's as a sorted tuple
+    cellIDsBySC = [ tuple(sorted(set(sc.values))) for sc in cn.subContours ]
+    # Get the ID's for all cells that contact the boundary, sorted
+    boundaryCells = sorted({ cid for cids in cellIDsBySC
+                                 if ( bgSet.intersection(cids) or
+                                      len(cids)==1 )
+                                 for cid in cids
+                                 if cid not in bgSet })
+    # Get all the real cell pairs (aka leave out the background)
+    cellPairs = sorted({ cids for cids in cellIDsBySC
+                              if not bgSet.intersection(cids) })
+    # Make a contact list for each cell
+    # Takes a cell ID as an index and returns all connecting cell ID's
+    cellContactDict = getElementConnections(cellPairs)
+    # Get all boundary cells with no neighbors besides other boundary cells
+    danglingCells = [ cid for cid in boundaryCells
+                          if not set(cellContactDict[cid]).difference(boundaryCells) ]
+    return danglingCells
+
+def GetCellNetworkWithCellsRemoved(cn,cellsToRemove,bgVals):
+    ''''Replace cells with holes in a copy of cn'''
+    cnClip = deepcopy(cn)
+    
+    # Make the new set of background values including the cells to remove
+    newBgSet = set(bgVals) | set(cellsToRemove)
+    
+    # Replace background-background interfaces with None
+    cnClip.subContours = [ ( sc if set(sc.values).difference(newBgSet) else
+                             None )
+                          for sc in cnClip.subContours ]
+    print [(0 if None else 1) for sc in cnClip.subContours]
+    # Delete cells from contourOrderingByValue and allValues and then re-index everything
+    print cellsToRemove
+    
+    for c in cellsToRemove:
+        del cnClip.contourOrderingByValue[c]
+    cnClip.allValues = [ v for v in cnClip.allValues if v not in cellsToRemove ]
+    cnClip.CleanUpEmptySubContours()
+    return cnClip
+
+def ClipDanglingBoundaryCells(cn,bgVals):
+    '''Remove all cells that only have neighbors that are also boundary cells
+       Returns the new CellNetwork and the new backround values
+         (which now include these cells)'''
+    cellsClip = GetDanglingBoundaryCells(cn,bgVals)
+    cnClip = GetCellNetworkWithCellsRemoved(cn,cellsClip,bgVals)
+    newBgVals = sorted(set(bgVals) | set(cellsClip))
+    return cnClip,newBgVals
+
 def GetCellNetworkListWithLimitedPointsBetweenNodes(cellNetworkList,splitLength=1,fixedNumInteriorPoints=None,interpolate=True,checkForDegeneracy=True):
     '''Based on matching subcontours by value pair, this function defines a fixed number of interior points for each subcontour
        and then applies this "trimming" procedure equitably to each frame in cellNetworkList (uses LimitPointsBetweenNodes)'''
@@ -623,6 +950,42 @@ def GetCellNetworkListWithLimitedPointsBetweenNodes(cellNetworkList,splitLength=
     
     return cellNetworkListNew
 
+def GetCellNetworkListWithLimitedPointsBetweenNodes_FullInterpolation(cellNetworkList,splitLength,checkForDegeneracy=True):
+    '''Based on matching subcontours by value pair, this function defines a fixed number of interior points for each subcontour
+       and then applies this "trimming" procedure equitably to each frame in cellNetworkList (uses BreakLinesIntoEvenPieces)'''
+    allPairs = sorted(set( [ tuple(sc.values) for cn in cellNetworkList for sc in cn.subContours ] )) # Value pairs...
+    
+    # Build the numInteriorPointsDict:
+    # minLength is the length of the shortest subcountour between cells p[0] and p[1] from all frames
+    minLength = { p : min( [ sum(pointDistance(sc.points[:-1],sc.points[1:]))
+                            for cn in cellNetworkList
+                            for sc in cn.subContours
+                            if tuple(sc.values)==p ] )
+                 for p in allPairs }
+    nSegmentsDict = { p:int(minLength[p]/splitLength) for p in allPairs }
+    
+    cellNetworkListNew = deepcopy(cellNetworkList) # otherwise, we'd also change the input argument in the outside world!
+    for cn in cellNetworkListNew:
+        for sc in cn.subContours:
+            sc.points = BreakLinesIntoEvenPieces(sc.points,nSegmentsDict[tuple(sc.values)])
+            sc.numPoints = len(sc.points)
+    
+    if checkForDegeneracy:
+        problemValsList,problemValuePairsList = ziptranspose([ cn.CheckForDegenerateContours() for cn in cellNetworkListNew ])
+        if flatten(problemValsList)!=[]:
+            print 'All degenerate values:',sorted(set(flatten(problemValsList)))
+            print 'Degenerate values by frame:'
+            for i,problemVals in enumerate(problemValsList):
+                if problemVals!=[]:
+                    print ' ',i,':',problemVals
+            print 'Degenerate contours (zero area) found between these cellIDs on these frames!'
+            for i,problemValuePairs in enumerate(problemValuePairsList):
+                if problemValuePairs!=[]:
+                    print ' ',i,':',problemValuePairs
+            raise DegenerateNetworkException('Degeneracy check failed!')
+    
+    return cellNetworkListNew
+
 def GetCellNetworkWithLimitedPointsBetweenNodes(cellNetwork,splitLength=1,fixedNumInteriorPoints=None,interpolate=True,checkForDegeneracy=True):
     '''Run GetCellNetworkListWithLimitedPointsBetweenNodes, but only for a single CellNetwork
        Useful if we don't want to limit points based on other frames
@@ -988,10 +1351,9 @@ def _loadCNListPrevNextFromJsonFile(cnListPrevAndNextFile):
 
 
 def GetMatchedCellNetworkListsPrevNext( waterArr,d,extraRemoveValsByFrame=None,forceRemake=False,
-                              bgVals=(0,1),scale=1,offset=(0,0), ):
+                                        bgVals=(0,1),scale=1,offset=(0,0), ):
     '''Get matched before and after CellNetwork lists from a waterArr.
        This function will optionally save and load to a pickle file (extraRemoveValsByFrame MUST be None)'''
-
     allValsByFrame = [ sorted( set(np.unique(i)).difference(bgVals) )
                       for i in waterArr ] # Skip background
     
@@ -1371,3 +1733,70 @@ def GetCellNetwork_NEW( watershed2d,allValues=None,bgVals=(0,1),scale=1,offset=(
     return CellNetwork( subContours=subContours,
                         contourOrderingByValue=contourOrderingByValue,
                         allValues=cellIDs )
+
+###########################################################################
+##    Some helper functions for dealing with (points,edges,cells) data   ##
+###########################################################################
+
+def _deIndexList(allVals,indexLoL):
+    return [ [ allVals[vid-1]
+              for vid in vIDs ]
+            for vIDs in indexLoL ]
+
+def _pointsFromEdgesByID(allPoints,edgesByPtID):
+    '''Takes a list of edges and returns a list of pairs of xy points'''
+    return _deIndexList(allPoints,edgesByPtID)
+
+def _getEdgesFromCellsByID(allEdges,cellsByEdgeID):
+    '''Takes a list of cells and returns a list of lists of edges'''
+    return _deIndexList(allEdges,cellsByEdgeID)
+
+def _getPointPairsFromCellsByID(allPointPairs,cellsByEdgeID):
+    '''Takes a list of cells and returns a list of lists of pairs of xy points'''
+    return _deIndexList(allPointPairs,cellsByEdgeID)
+
+def getPointsEdgesCellsFromCellNetwork(cn,arr):
+    allValues = np.unique(arr)
+    allPoints = sorted(set(cn.GetAllPoints()))
+    pindDict = { pt:(i+1) for i,pt in enumerate(allPoints) }
+    # Create an "enhanced" edge list that has elements like:
+    # (sorted point indices, cell values)
+    allPtIDsAndValues = [ ( sorted( pindDict[tuple(p)]
+                                   for p in [p1,p2] ),
+                            sc.values )
+                         for sc in cn.subContours
+                         for p1,p2 in zip(sc.points[:-1],sc.points[1:]) ]
+    allEdgesAsPtIDs = sorted({ tuple(pids) for pids,cellIDs in allPtIDsAndValues })
+    eindDict = { tuple(edge):(i+1) for i,edge in enumerate(allEdgesAsPtIDs) }
+    allCellsAsEdgeIDs = [ sorted({ eindDict[tuple(pids)]
+                                 for pids,cellIDs in allPtIDsAndValues
+                                 if c in cellIDs })
+                        for c in allValues ]
+    orderedPointsByCell = { cellID:removeDuplicates(cn.GetContourPoints(cellID))
+                           for cellID in cn.contourOrderingByValue.keys() }
+    orderedPointIDsByCell = { k:[ pindDict[tuple(i)]
+                                 for i in v ]
+                             for k,v in orderedPointsByCell.items() }
+    pointPairsByCell = { k:map(tuple,map(sorted,zip(v,roll(v))))
+                        for k,v in orderedPointIDsByCell.items() }
+    edgesByCell = [ [k,[ eindDict[i]
+                       for i in v
+                       if i in eindDict ]]
+                   for k,v in pointPairsByCell.items() ]
+    BAD_PAIRS = [ [allPoints[i[0]-1],allPoints[i[1]-1]]
+                 for k,v in pointPairsByCell.items()
+                 for i in v
+                 if i not in eindDict ]
+    print BAD_PAIRS, 'Not sure why these are here... ignored for the time being!'
+    
+    return allPoints,allEdgesAsPtIDs,allCellsAsEdgeIDs
+
+def getTripleJunctions(points,edges):
+    pointCounts = Counter( flatten(edges,2) )
+    pointCountBins = sorted(edgeCounts.values())
+    pointsByOccurrence = { c:sorted( k for k,v in edgeCounts.items() 
+                                       if v==c )
+                          for c in pointCountBins }
+    nonJunctionPoints = set(points[i-1] for i in pointsByOccurrence[2])
+    junctionPoints = sorted(set(points).difference(nonJunctionPoints))
+    return junctionPoints
